import fs from 'fs'
import { BaseEvent } from '@/application/base/baseEvent'
import { SocketEvent } from '@/application/enums'
import type { UUID } from '@/application/types'
import { SpriteAction } from '@/entities/spriteAction'
import SpriteRepository from '@/repositories/spriteRepository'
import sharp from 'sharp'

interface SpriteImage {
  url: string
  offset: {
    x: number
    y: number
  }
}

interface ImageDimensions {
  width: number
  height: number
  offsetX: number
  offsetY: number
}

interface EffectiveDimensions {
  width: number
  height: number
  top: number
  bottom: number
}

interface SpriteActionData {
  action: string
  sprites: SpriteImage[]
  originX: number
  originY: number
  frameRate: number
}

interface UpdateSpritePayload {
  id: UUID
  name: string
  spriteActions: SpriteActionData[]
}

export default class SpriteUpdateEvent extends BaseEvent {
  private readonly spriteRepository: SpriteRepository = new SpriteRepository()

  public listen(): void {
    this.socket.on(SocketEvent.GM_SPRITE_UPDATE, this.handleEvent.bind(this))
  }

  private async handleEvent(data: UpdateSpritePayload, callback: (success: boolean) => void): Promise<void> {
    try {
      if (!(await this.isCharacterGM())) {
        callback(false)
        return
      }

      const sprite = await this.spriteRepository.getById(data.id)
      if (!sprite) {
        callback(false)
        return
      }

      await this.spriteRepository.getEntityManager().populate(sprite, ['spriteActions'])

      // Update sprite in database
      await sprite.setName(data.name).setUpdatedAt(new Date()).save()

      // First verify all sprite sheets can be generated
      const allSheetsGenerated = await this.verifyAllSpriteSheets(data.spriteActions, sprite.getId())
      if (!allSheetsGenerated) {
        callback(false)
        return
      }

      // Remove existing actions
      await this.removeExistingActions(sprite)

      // Create new actions
      await this.createNewSpriteActions(data.spriteActions, sprite)

      callback(true)
    } catch (error) {
      console.error(`Error updating sprite ${data.id}:`, error)
      callback(false)
    }
  }

  private async verifyAllSpriteSheets(actionsData: SpriteActionData[], spriteId: string): Promise<boolean> {
    for (const actionData of actionsData) {
      if (!(await this.generateSpriteSheet(actionData.sprites, spriteId, actionData.action))) {
        return false
      }
    }
    return true
  }

  private async removeExistingActions(sprite: any): Promise<void> {
    const existingActions = sprite.getSpriteActions()
    for (const existingAction of existingActions) {
      await this.spriteRepository.getEntityManager().removeAndFlush(existingAction)
    }
  }

  private async createNewSpriteActions(actionsData: SpriteActionData[], sprite: any): Promise<void> {
    for (const actionData of actionsData) {
      // Process images and calculate dimensions
      const imageData = await Promise.all(actionData.sprites.map((sprite) => this.processImage(sprite)))
      const effectiveDimensions = imageData.map((dimensions) => this.calculateEffectiveDimensions(dimensions))

      // Calculate total height needed for the sprite sheet
      const maxHeight = Math.max(...effectiveDimensions.map((d) => d.height), 0)
      const maxTop = Math.max(...effectiveDimensions.map((d) => d.top), 0)
      const maxBottom = Math.max(...effectiveDimensions.map((d) => d.bottom), 0)
      const totalHeight = maxHeight + maxTop + maxBottom

      const spriteAction = new SpriteAction()
      spriteAction.setSprite(sprite)
      sprite.getSpriteActions().add(spriteAction)

      const maxWidth = await this.calculateMaxWidth(actionData.sprites)

      spriteAction
        .setAction(actionData.action)
        .setSprites(actionData.sprites)
        .setOriginX(actionData.originX)
        .setOriginY(actionData.originY)
        .setFrameWidth(maxWidth)
        .setFrameHeight(totalHeight)
        .setFrameRate(actionData.frameRate)

      await this.spriteRepository.getEntityManager().persistAndFlush(spriteAction)
    }
  }

  private async generateSpriteSheet(sprites: SpriteImage[], spriteId: string, action: string): Promise<boolean> {
    try {
      if (!sprites.length) return true

      // Process all images and get their dimensions
      const imageData = await Promise.all(sprites.map((sprite) => this.processImage(sprite)))
      const effectiveDimensions = imageData.map((dimensions) => this.calculateEffectiveDimensions(dimensions))

      // Calculate maximum dimensions
      const maxWidth = Math.max(...effectiveDimensions.map((d) => d.width), 0)
      const maxHeight = Math.max(...effectiveDimensions.map((d) => d.height), 0)
      const maxTop = Math.max(...effectiveDimensions.map((d) => d.top), 0)
      const maxBottom = Math.max(...effectiveDimensions.map((d) => d.bottom), 0)

      // Calculate total height needed
      const totalHeight = maxHeight + maxTop + maxBottom

      // Process images and create sprite sheet
      const processedImages = await Promise.all(
        sprites.map(async (sprite) => {
          const { width, height, offsetX, offsetY } = await this.processImage(sprite)
          const uri = this.extractBase64Data(sprite.url)
          const buffer = Buffer.from(uri, 'base64')

          // Create individual frame
          const left = offsetX >= 0 ? offsetX : 0
          const verticalOffset = totalHeight - height - (offsetY >= 0 ? offsetY : 0)
          return sharp({
            create: {
              width: maxWidth,
              height: totalHeight,
              channels: 4,
              background: { r: 0, g: 0, b: 0, alpha: 0 }
            }
          })
            .composite([{ input: buffer, left, top: verticalOffset }])
            .png()
            .toBuffer()
        })
      )

      // Combine frames into sprite sheet
      const spriteSheet = await this.createSpriteSheetImage(processedImages, maxWidth, totalHeight)

      // Save the sprite sheet
      await this.saveSpriteSheet(spriteId, action, spriteSheet)

      return true
    } catch (error) {
      console.error('Error generating sprite sheet:', error)
      return false
    }
  }

  private async createSpriteSheetImage(processedImages: Buffer[], maxWidth: number, totalHeight: number): Promise<Buffer> {
    return sharp({
      create: {
        width: maxWidth * processedImages.length,
        height: totalHeight,
        channels: 4,
        background: { r: 0, g: 0, b: 0, alpha: 0 }
      }
    })
      .composite(
        processedImages.map((buffer, index) => ({
          input: buffer,
          left: index * maxWidth,
          top: 0
        }))
      )
      .png()
      .toBuffer()
  }

  private async saveSpriteSheet(spriteId: string, action: string, spriteSheet: Buffer): Promise<void> {
    // Ensure directory exists
    const dir = `public/sprites/${spriteId}`
    await fs.promises.mkdir(dir, { recursive: true })

    // Save the sprite sheet
    await fs.promises.writeFile(`${dir}/${action}.png`, spriteSheet)
  }

  private extractBase64Data(url: string): string {
    const uri = url.split(';base64,').pop()
    if (!uri) throw new Error('Invalid base64 image')
    return uri
  }

  private async processImage(sprite: SpriteImage): Promise<ImageDimensions> {
    const uri = this.extractBase64Data(sprite.url)
    const buffer = Buffer.from(uri, 'base64')
    const metadata = await sharp(buffer).metadata()
    return {
      width: metadata.width ?? 0,
      height: metadata.height ?? 0,
      offsetX: sprite.offset?.x ?? 0,
      offsetY: sprite.offset?.y ?? 0
    }
  }

  private calculateEffectiveDimensions(imageDimensions: ImageDimensions): EffectiveDimensions {
    return {
      width: imageDimensions.width + Math.abs(imageDimensions.offsetX),
      height: imageDimensions.height + Math.abs(imageDimensions.offsetY),
      top: imageDimensions.offsetY >= 0 ? imageDimensions.offsetY : 0,
      bottom: imageDimensions.offsetY < 0 ? Math.abs(imageDimensions.offsetY) : 0
    }
  }

  private async calculateMaxWidth(sprites: SpriteImage[]): Promise<number> {
    if (!sprites.length) return 0

    // Process all images and get their dimensions
    const imageData = await Promise.all(sprites.map((sprite) => this.processImage(sprite)))
    const effectiveDimensions = imageData.map((dimensions) => this.calculateEffectiveDimensions(dimensions))

    // Calculate maximum width needed
    return Math.max(...effectiveDimensions.map((d) => d.width), 0)
  }
}
